/* **
    Monitor de disponibilidad de servicios
    Gabriel Ahumada     gabriel.ahumada@alumnos.usm.cl
    Constanza Valdés    constanza.valdes@alumnos.usm.cl  
* **/

#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/wait.h>
#include <signal.h>
#include <sys/time.h>

#define SIZE 200

char auth[SIZE];
char smtp[SIZE];
char mail[SIZE];

/* Función que envía el correo al administrador */
void sendEmail(int action){

    char fail[SIZE];
    char mail_aux[SIZE];
    pid_t pid;
    int pfdToChild[2];
    int pfdFromChild[2];
    int status;
    int port = 587;
    FILE * toChild, *fromChild;
    char line[SIZE];

    /* Clasificación del mensaje a enviar según el tipo de error */
    switch(action){
        case 1: strcpy(fail,"DNS doesn't work!!"); printf("DNS fail\n"); break;
        case 2: strcpy(fail,"Website is down!!"); printf("Website is down\n"); break;
        case 3: strcpy(fail,"CPU is overload!!"); printf("CPU is overload\n"); break;
        case 4: strcpy(fail,"Memory is overload!!"); printf("Memory is overload\n"); break;
        default: printf("Unknown error");
    }    

    /* Creación de pipe para la comunicación entre procesos */
    if(pipe(pfdToChild) < 0){
        perror("pipe");
        exit(1);
    }
    if(pipe(pfdFromChild) < 0){
        perror("pipe");
        exit(1);
    }

    /* Creación de proceso hijo */
    if((pid = fork()) < 0){
        perror("fork");
        exit(1);
    }

    /* Código ejecutado por el hijo */
    if(pid == 0){
      dup2(pfdToChild[0], 0);
      close(pfdToChild[1]);
      
      dup2(pfdFromChild[1],1);
      dup2(pfdFromChild[1],2);
      close(pfdFromChild[0]);
      
      sprintf(line,"%s:%i",smtp,port);

      /* Comando openssl utilizado para conectarse al cliente de correos */
      execl("/usr/bin/openssl", "openssl", "s_client",
            "-starttls","smtp", "-crlf", "-connect", line, NULL);
      perror("exec");
      _exit(127);
    }

    /* Código ejecutado por el padre */
    close(pfdToChild[0]);
    close(pfdFromChild[1]);
    
    toChild = fdopen(pfdToChild[1], "w");
    fromChild = fdopen(pfdFromChild[0], "r");
    sprintf(mail_aux,"<%s>",mail);

    /* Formación del correo a enviar por la pipe */
    while (strncmp(fgets(line, sizeof(line), fromChild),"250 ",4)!=0);
    
    fprintf(toChild,"AUTH PLAIN %s\n", auth);
    fflush(toChild);
    fgets(line, sizeof(line), fromChild);

    fprintf(toChild,"mail from: %s\n",mail_aux);
    fflush(toChild);
    fgets(line, sizeof(line), fromChild);

    fprintf(toChild,"rcpt to: %s\n",mail_aux);
    fflush(toChild);
    fgets(line, sizeof(line), fromChild);

    fprintf(toChild,"data\n");
    fflush(toChild);
    fgets(line, sizeof(line), fromChild);

    fprintf(toChild, "From:\" - \" %s \n To: \"ME\" <me@usm.cl>\nSubject: Alert!\n\n   %s  .\n.\n",mail_aux,fail);
    fflush(toChild);
    fgets(line, sizeof(line), fromChild);

    fprintf(toChild,"quit\n");
    fflush(toChild);
    fgets(line, sizeof(line), fromChild);

    /* Cierre de pipe y espera la culminación del hijo */
    fclose(toChild);
    fclose(fromChild);
    waitpid(pid, &status, 0);
}

/* Función que monitorea el estado de un DNS */
void checkDNS(char* argv){

    pid_t pid;
    int pfdtoChild[2];
    int pfdfromChild[2];    
    int status;
    char line[256];
    char DNS[10] = "200.1.17.6";
    char DNS_ans[11];
    int i=0;
    int aux=0;
    int j=0;

    /* Creación de pipe para la comunicación entre procesos */
    if(pipe(pfdtoChild) < 0){
        perror("pipe");
        exit(1);
    }
    if(pipe(pfdfromChild) < 0){
        perror("pipe");
        exit(1);
    }

    /* Creación de proceso hijo */
    if((pid=fork()) < 0){
        perror("fork");
        exit(1);
    }

    /* Código ejecutado por el hijo */
    if(pid == 0){
        dup2(pfdtoChild[0], 0);
        dup2(pfdfromChild[1], 1);

        close(pfdtoChild[1]);
        close(pfdfromChild[0]);

        fflush(stdin);

        /* Comando nslookup utilizado para obtener el encabezado de la página web */
        execl("/usr/bin/nslookup", "nslookup", "-", argv, NULL);        
        perror("exec");
        _exit(127);
    }

     /* Código ejecutado por el padre */
    close(pfdtoChild[0]);
    close(pfdfromChild[1]);
    
    /* Envío de consulta a DNS */
    write(pfdtoChild[1],"profesores.elo.utfsm.cl\n",24);
    
    /* Obtención de la salida del comando ejecutado por el hijo */
    read(pfdfromChild[0], line,256);
    for(i = 0;i < 256;i++){
        if(line[i] == 'r' && line[i+1] == 'e' && line[i+2] == 's' && line[i+3] == 's' && 
            line[i+4] == ':' && line[i+5] == ' ')
            aux++;
        if(aux == 1){
            i= i+6;
            break;
        }
    }

    while(line[i+j] != '\n') {
        DNS_ans[j] = line[i+j];
        j++;
    }
    DNS_ans[j] = '\0';

    /* Si la respuesta no corresponde a la esperada
    se invoca la función para enviar correo */
    if(strncmp(DNS,DNS_ans,10)==0) printf("DNS is OK\n");
    else sendEmail(1);

    /* Cierre de pipe y espera la culminación del hijo */
    close(pfdtoChild[1]);
    close(pfdfromChild[1]);
    waitpid(pid, &status, 0);
}

/* Función que monitorea el estado de una página web */
void checkWebsite(char* argv){

    pid_t pid;
    int pfdtoChild[2];
    int pfdfromChild[2];    
    int status;
    char line[256];
    char code_okn[10]="HTTP/1.1 2";
    char code_oko[10]="HTTP/1.0 2";
    char code_pgn[10]="HTTP/1.1 3";
    char code_pgo[10]="HTTP/1.0 3";

    /* Creación de pipe para la comunicación entre procesos */
    if(pipe(pfdtoChild) < 0){
        perror("pipe");
        exit(1);
    }
    if(pipe(pfdfromChild) < 0){
        perror("pipe");
        exit(1);
    }

    /* Creación de proceso hijo */
    if((pid=fork()) < 0){
        perror("fork");
        exit(1);
    }

    /* Código ejecutado por el hijo */
    if (pid == 0) {
        dup2(pfdtoChild[0], 0);
        dup2(pfdfromChild[1], 1);
        close(pfdtoChild[1]);
        close(pfdfromChild[0]);

        fflush(stdin);

        /* Comando curl utilizado para obtener el encabezado de la página web */
        execl("/usr/bin/curl", "curl", "-s", "--head", argv, NULL);
        perror("exec");
        _exit(127);
    }
    
    /* Código ejecutado por el padre */
    close(pfdtoChild[0]);
    close(pfdfromChild[1]);

    /* Obtención de la salida del comando ejecutado por el hijo */
    read(pfdfromChild[0], line,20);

    /* Si el código HTTP retornado no es 2XX o 3XX 
    se invoca la función para enviar correo */
    if(strncmp(code_okn,line,10)==0 || strncmp(code_oko,line,10)==0 
        || strncmp(code_pgn,line,10)==0 || strncmp(code_pgo,line,10)==0)
        printf("Website is OK\n");
    else sendEmail(2);

    /* Cierre de pipe y espera la culminación del hijo */
    close(pfdtoChild[1]);
    close(pfdfromChild[1]);
    waitpid(pid, &status, 0);
}

/* Función que monitorea la cantidad de CPU y memoria utilizada */
void checkPerformance(char *argv_cpu, char *argv_mem){

    pid_t pid;
    int pfdtoChild[2];
    int pfdfromChild[2];    
    int status;
    char cpu[4];
    char mem[4];
    float cpu_p=0;
    float mem_p=0;
    FILE * temp;
    int c;
    
    /* Creación de pipe para la comunicación entre procesos */    
    if(pipe(pfdtoChild) < 0){
        perror("pipe");
        exit(1);
    }
    if(pipe(pfdfromChild) < 0){
        perror("pipe");
        exit(1);
    }

    /* Creación de proceso hijo */
    if((pid=fork()) < 0){
        perror("fork");
        exit(1);
    }

    /* Código ejecutado por el hijo */
    if (pid == 0) {
        dup2(pfdtoChild[0], 0);
        dup2(pfdfromChild[1], 1);
        close(pfdtoChild[1]);
        close(pfdfromChild[0]);

        fflush(stdin);

        /* Comando ps utilizado para obtener el uso de CPU 
        y memoria de todos los procesos */
        execl("/bin/ps", "ps", "-eo", "pcpu,pmem", NULL);
        perror("exec");
        _exit(127);
    }
    
    /* Código ejecutado por el padre */
    close(pfdtoChild[0]);
    close(pfdfromChild[1]);

    /* Obtención de la salida del comando ejecutado por el hijo */
    temp=fdopen(pfdfromChild[0],"r");
    while((c=fgetc(temp))!=EOF){
        fscanf(temp,"%s %s",cpu,mem);
        cpu_p+=atof(cpu);
        mem_p+=atof(mem);
    }
    fclose(temp);
    
    /* Si la suma de utilización de CPU y memoria supera a los datos 
    entregados por el usuario se invoca la función para enviar correo */
    if(cpu_p>atof(argv_cpu)) sendEmail(3);
    else printf("CPU is OK\n");
    if(mem_p>atof(argv_mem)) sendEmail(4);
    else printf("Memory is OK\n");

    /* Cierre de pipe y espera la culminación del hijo */
    close(pfdtoChild[1]);
    close(pfdfromChild[1]);
    waitpid(pid, &status, 0);
}


/* Función principal */
int main (int argc, char* argv[]){

    /* Revisión del número de argumentos ingresados */
    if(argc != 8){
        perror("Error in the number of arguments");
        exit(1);
    }  

    /* Estructura con el contador de tiempo */
    struct itimerval contador;

    /* Valor inicial del contador */
    struct timeval tiempoInicial;

    /* Tiempo de repetición del contador */
    struct timeval tiempoRepeticion;

    /* Se rellena el tiempo inicial del contador con 1 segundos */
    tiempoInicial.tv_sec=1;
    tiempoInicial.tv_usec=0;

    /* Se rellena el tiempo de repetición con 60 segundo */
    tiempoRepeticion.tv_sec=60;
    tiempoRepeticion.tv_usec=0;

    /* Se rellenan los datos del contador */
    contador.it_value=tiempoInicial;
    contador.it_interval=tiempoRepeticion;

    sigset_t mask;
    int signo;

    setitimer(ITIMER_REAL, & contador, NULL);

    /* Asignación de señales de alarma */
    sigemptyset(&mask);
    sigaddset(&mask,SIGALRM);
    sigprocmask(SIG_BLOCK, &mask, NULL);

    /* Adquisición de los datos para el envío de correo */
    strcpy(smtp, argv[5]);
    strcpy(mail, argv[6]);
    strcpy(auth, argv[7]);


    while (1){
        /* Diferente a sigalarm */
        signo=SIGINT;  
        /* Espera por la señal de alarma */
        while(signo!=SIGALRM) sigwait(&mask, &signo); 

        /* Monitoreo de los servicios disponibles */
        checkDNS(argv[1]);
        checkWebsite(argv[2]);
        checkPerformance(argv[3],argv[4]);
    }
    exit(0);
}
